from __future__ import print_function
import numpy as np
from scipy import spatial
from . import wcsutils, utils, enmap, coordinates, fft, curvedsky

# Python 2/3 compatibility
try: basestring
except NameError: basestring = str

def thumbnails(imap, coords, r=5*utils.arcmin, res=None, proj=None, apod=2*utils.arcmin,
		method="mixed", order=3, oversample=4, pol=None, oshape=None, owcs=None,
		extensive=False, verbose=False, filter=None,pixwin=False,pixwin_order=0):
	"""Given an enmap [...,ny,nx] and a set of coordinates in a numpy array
	coords with shape (n,2) and ordering [n,{dec,ra}], extract a set
	of thumbnail images [n,...,thumby,thumbx] centered on each set of
	coordinates, where n is the number of images. 
	When proj is 'tan', each of these thumbnail images is projected onto a local 
	tangent plane, removing the effect of size and shape distortions in the input 
	map.

	If oshape, owcs are specified, then the thumbnails will have this geometry,
	which should be centered on [0,0]. Otherwise, a geometry with the given
	projection (defaults to "tan" = gnomonic projection) will be constructed,
	going up to a maximum radius of r. FIXME: Defaults to "car" instead while
	enmap.pixsizemap is buggy for non-cylindrical projections.

	The reprojection involved in this operation implies interpolation. The default
	is to use fft rescaling to oversample the input pixels by the given pixel, and
	then use bicubic spline interpolation to read off the values at the output
	pixel centers. The fft oversampling can be controlled with the oversample argument.
	Values <= 1 turns this off. The other interpolation step is controlled using the
	"order" argument. 0/1/3 corresponds to nearest neighbor, bilinear and bicubic spline
	interpolation respectively.

	If pol == True, then Q,U will be rotated to take into account the change in
	the local northward direction impled in the reprojection. The default is to
	do polarization rotation automatically if the input map has a compatible shape,
	e.g. at least 3 axes and a length of 3 for the 3rd last one. TODO: I haven't
	tested this yet.

	If extensive == True (not the default), then the map is assumed to contain an
	extensive field rather than an intensive one. An extensive field is one where
	the values in the pixels depend on the size of the pixel. For example, if the
	inverse variance in the map is given per pixel, then this ivar map will be
	extensive, but if it's given in units of inverse variance per square arcmin
	then it's intensive.

	For reprojecting inverse variance maps, consider using the wrapper thumbnails_ivar,
	which makes it easier to avoid common pitfalls.
	
	If pixwin is True, the pixel window will be deconvolved."""
	# FIXME: Specifying a geometry manually is broken - see usage of r in neighborhood_pixboxes below
	# Handle arbitrary coords shape
	if proj is None: proj = "car"
	coords = np.asarray(coords)
	ishape = coords.shape[:-1]
	coords = coords.reshape(-1, coords.shape[-1])
	# If the output geometry was not given explicitly, then build one
	if oshape is None:
		if res is None: res = min(np.abs(imap.wcs.wcs.cdelt))*utils.degree/2
		oshape, owcs = enmap.thumbnail_geometry(r=r, res=res, proj=proj)
	# Check if we should be doing polarization rotation
	pol_compat = imap.ndim >= 3 and imap.shape[-3] == 3
	if pol is None: pol = pol_compat
	if pol and not pol_compat: raise ValueError("Polarization rotation requested, but can't interpret map shape %s as IQU map" % (str(imap.shape)))
	nsrc = len(coords)
	if verbose: print("Extracting %d %dx%d thumbnails from %s map" % (nsrc, oshape[-2], oshape[-1], str(imap.shape)))
	opos = enmap.posmap(oshape, owcs)
	# Get the pixel area around each of the coordinates
	rtot     = r + apod
	apod_pix = utils.nint(apod/(np.min(np.abs(imap.wcs.wcs.cdelt))*utils.degree))
	pixboxes = enmap.neighborhood_pixboxes(imap.shape, imap.wcs, coords, rtot)
	# Define our output maps, which we will fill below
	omaps = enmap.zeros((nsrc,)+imap.shape[:-2]+oshape, owcs, imap.dtype)
	for si, pixbox in enumerate(pixboxes):
		if method == "nufft" or method == "mixed" and oversample > 1:
			# Make the pixbox fft-friendly
			for i in range(2):
				pixbox[1,i] = pixbox[0,i] + fft.fft_len(pixbox[1,i]-pixbox[0,i], direction="above", factors=[2,3,5])
		ithumb = imap.extract_pixbox(pixbox)
		if extensive: ithumb /= ithumb.pixsizemap()
		ithumb = ithumb.apod(apod_pix, fill="median")
		if pixwin: ithumb = enmap.unapply_window(ithumb, order=pixwin_order)
		if filter is not None: ithumb = filter(ithumb)
		if verbose:
			print("%4d/%d %6.2f %6.2f %8.2f %dx%d" % (si+1, nsrc, coords[si,0]/utils.degree, coords[si,1]/utils.degree, np.max(ithumb), ithumb.shape[-2], ithumb.shape[-1]))
		# Oversample using fourier if requested. We do this because fourier
		# interpolation is better than spline interpolation overall
		if method == "mixed" and oversample > 1:
			fshape = utils.nint(np.array(oshape[-2:])*oversample)
			ithumb = ithumb.resample(fshape, method="fft")
		# I apologize for the syntax. There should be a better way to do this
		ipos = coordinates.transform("cel", ["cel",[[0,0,coords[si,1],coords[si,0]],False]], opos[::-1], pol=pol)
		ipos, rest = ipos[1::-1], ipos[2:]
		if   method in ["spline", "mixed"]: ipol_mode = "spline"
		elif method in ["nufft"]:           ipol_mode = "nufft"
		else: raise ValueError("Unrecognized method '%s'" % (str(method)))
		omaps[si] = ithumb.at(ipos, mode=ipol_mode, order=order)
		# Apply the polarization rotation. The sign is flipped because we computed the
		# rotation from the output to the input
		if pol: omaps[si] = enmap.rotate_pol(omaps[si], -rest[0])
	if extensive: omaps *= omaps.pixsizemap()
	# Restore original dimension
	omaps = omaps.reshape(ishape + omaps.shape[1:])
	return omaps

def thumbnails_ivar(imap, coords, r=5*utils.arcmin, res=None, proj=None,
		oshape=None, owcs=None, order=1, extensive=True, verbose=False):
	"""Like thumbnails, but for hitcounts, ivars, masks, and other quantities that
	should stay positive and local. Remember to set extensive to True if you have an
	extensive quantity, i.e. if the values in each pixel would go up if multiple pixels
	combined. An example of this is a hitcount map or ivar per pixel. Conversely, if
	you have an intensive quantity like ivar per arcmin you should set extensive=False."""
	return thumbnails(imap, coords, r=r, res=res, proj=proj, oshape=oshape, owcs=owcs,
			order=order, oversample=1, pol=False, extensive=extensive, verbose=verbose,
			pixwin=False)

def map2healpix(imap, nside=None, lmax=None, out=None, rot=None, spin=[0,2], method="harm", order=1, extensive=False, bsize=100000, nside_mode="pow2", boundary="constant", verbose=False, niter=0):
	"""Reproject from an enmap to healpix, optionally including a rotation.

	imap:  The input enmap[...,ny,nx]. Stokes along the -3rd axis if
	       present.
	nside: The nside of the healpix map to generate. Not used if
	       an output map is passed. Otherwise defaults to the same
	       resolution as the input map.
	lmax:  The highest multipole to use in any harmonic-space
	       operations. Defaults to the input maps' Nyquist limit.
	out:   An optional array [...,npix] to write the output map to.
	       The ...  part must match the input map, as must the data
	       type.
	rot:   An optional coordinate rotation to apply. Either a string
	       "isys,osys", where isys is the system to transform from,
	       and osys is the system to transform to. Currently the values
	       "cel"/"equ" and "gal" are recognized. Alternatively, a tuple of
	       3 euler zyz euler angles can be passed, in the same convention
	       as healpy.rotate_alm.
	spin:  A description of the spin of the entries along the stokes
	       axis. Defaults to [0,2], which means that the first entry
	       is spin-0, followed by a spin-2 pair (any non-zero spin
	       covers a pair of entries). If the axis is longer than
	       what's covered in the description, then it is repeated as
	       necessary. Pass spin=[0] to disable any special treatment
	       of this axis.
	method: How to interpolate between the input and output
	       pixelizations. Can be "harm" (default) or "spline".
	       "harm" maps between them using spherical harmonics
	       transforms. This preserves the power spectrum (so no window
	       function is introduced), and averages noise down when the
	       output pixels are larger than the input pixels. However, it
	       can suffer from ringing around very bright features, and an
	       all-positive input map may end up with small negative values.
	       "spline" instead uses spline interpolation to look up the
	       value in the intput map corresponding to each pixel center
	       in the output map. The spline order is controlled with the
	       "order" argument. Overall "harm" is best suited for normal
	       sky maps, while "spline" with order = 0 or 1 is best suited
	       for hitcount maps and masks.
	order: The spline order to use when method="spline".
	       0 corresponds to nearest neighbor interpolation.
	       1 corresponds to bilinear interpolation (default)
	       3 corresponds to bicubic spline interpolation.
	       0 and 1 are local and do not introduce values outside
	       the input range, but introduce some aliasing and loss of
	       power. 3 has less power loss, but still non-zero, and
	       is vulnerable to ringing.
	extensive: Whether the map represents an extensive (as opposed to
	       intensive) quantity. Extensive quantities have values
	       proportional to the pixel size, unlike intensive quantities.
	       Hitcount per pixel is an extensive quantity. Hitcount per
	       square degree is an intensive quantity, as is a temperature
	       map. Defaults to False.
	bsize: The spline method operates on batches of pixels to save memory.
	       This controls the batch size, in pixels. Defaults to 100000.
	nside_mode: Controls which restrictions apply to nside in the case where
	       it has to be inferred automatically. Can be "pow2", "mul32" and "any".
	       "pow2", the default, results in nside being a power of two, as
	       required by the healpix standard.
	       "mul32" relaxes this requirement, making a map where nside is a
	       multiple of 32. This is compatible with most healpix operations,
	       but not with ud_grade or the nest pixel ordering.
	       "any" allows for any integer nside.
	boundary: The boundary conditions assumed for the input map when
	       method="spline". Defaults to "constant", which assumes that
	       anything outsize the map has a constant value of 0. Another
	       useful value is "wrap", which assumes that the right side
	       wraps over to the left, and the top to the bottom. See
	       scipy.ndimage.distance_transform's documentation for other,
	       less useful values. method="harm" always assumes "constant"
	       regardless of this setting.
	verbose: Whether to print information about what it's doing.
	       Defaults to False, which doesn't print anything.

	Typical usage:
	* map_healpix  = map2healpix(map,  rot="cel,gal")
	* ivar_healpix = map2healpix(ivar, rot="cel,gal", method="spline", spin=[0], extensive=True)
	"""
	# Get the map's typical resolution from cdelt
	ires = np.mean(np.abs(imap.wcs.wcs.cdelt))*utils.degree
	lnyq = np.pi/ires
	if out is None:
		if nside is None:
			nside = restrict_nside(((4*np.pi/ires**2)/12)**0.5, nside_mode)
		out = np.zeros(imap.shape[:-2]+(12*nside**2,), imap.dtype)
	npix     = out.shape[-1]
	opixsize = 4*np.pi/npix
	# Might not be safe to go all the way to the Nyquist l, but looks that way to my tests.
	if lmax is None: lmax = lnyq
	if extensive:
		imap = imap * (opixsize / imap.pixsizemap(broadcastable=True)) # not /= to avoid changing original imap
	if method in ["harm", "harmonic"]:
		# Harmonic interpolation preserves the power spectrum, but can introduce ringing.
		# Probably not a good choice for positive-only quantities like hitcounts.
		# Coordinate rotation is slow.
		alm = curvedsky.map2alm(imap, lmax=lmax, spin=spin, niter=niter)
		if rot is not None:
			curvedsky.rotate_alm(alm, *rot2euler(rot), inplace=True)
		curvedsky.alm2map_healpix(alm, out, spin=spin)
		del alm
	elif method == "spline":
		# Covers both cubic spline interpolation (order=3), linear interpolation (order=1)
		# and nearest neighbor (order=0). Harmonic interpolation is preferable to cubic
		# splines, but linear and nearest neighbor may be useful. Coordinate rotation may
		# be slow.
		import healpy
		ip = utils.interpolator(imap, npre=-2, order=order, mode="spline", border=boundary)
		# Figure out if we need to compute polarization rotations
		pol = imap.ndim > 2 and any([s != 0 for s,c1,c2 in enmap.spin_helper(spin, imap.shape[-3])])
		# Batch to save memory
		for i1 in range(0, npix, bsize):
			i2   = min(i1+bsize, npix)
			opix = np.arange(i1,i2)
			pos  = healpy.pix2ang(nside, opix)[::-1]
			pos[1][:] = np.pi/2-pos[1]
			if rot is not None:
				# Not sure why the [::-1] is necessary here. Maybe psi,theta,phi vs. phi,theta,psi?
				pos = coordinates.transform_euler(inv_euler(rot2euler(rot))[::-1], pos, pol=pol)
			# The actual interpolation happens here
			vals  = imap.at(pos[1::-1], order=order, border=boundary, ip=ip)
			if rot is not None and imap.ndim > 2:
				# Update the polarization to account for the new coordinate system
				for s, c1, c2 in enmap.spin_helper(spin, imap.shape[-3]):
					if s != 0:
						vals = enmap.rotate_pol(vals, -pos[2], spin=s, comps=[c1,c2-1], axis=-2)
			out[...,i1:i2] = vals
	else:
		raise ValueError("Map reprojection method '%s' not recognized" % str(method))
	return out

def healpix2map(iheal, shape=None, wcs=None, lmax=None, out=None, rot=None, spin=[0,2], method="harm", order=1, extensive=False, bsize=100000, verbose=False, niter=0):
	"""Reproject from healpix to an enmap, optionally including a rotation.

	iheal: The input healpix map [...,npix]. Stokes along the -2nd axis if
	       present.
	shape: The (...,ny,nx) shape of the output map. Only the last two entries
	       are used, the rest of the dimensions are taken from iheal.
	       Mandatory unless an output map is passed.
	wcs  : The world woordinate system object the output map.
	       Mandatory unless an output map is passed.
	lmax:  The highest multipole to use in any harmonic-space
	       operations. Defaults to 3 times the nside of iheal.
	out:   An optional enmap [...,ny,nx] to write the output map to.
	       The ...  part must match iheal, as must the data type.
	rot:   An optional coordinate rotation to apply. Either a string
	       "isys,osys", where isys is the system to transform from,
	       and osys is the system to transform to. Currently the values
	       "cel"/"equ" and "gal" are recognized. Alternatively, a tuple of
	       3 euler zyz euler angles can be passed, in the same convention
	       as healpy.rotate_alm.
	spin:  A description of the spin of the entries along the stokes
	       axis. Defaults to [0,2], which means that the first entry
	       is spin-0, followed by a spin-2 pair (any non-zero spin
	       covers a pair of entries). If the axis is longer than
	       what's covered in the description, then it is repeated as
	       necessary. Pass spin=[0] to disable any special treatment
	       of this axis.
	method: How to interpolate between the input and output
	       pixelizations. Can be "harm" (default) or "spline".
	       "harm" maps between them using spherical harmonics
	       transforms. This preserves the power spectrum (so no window
	       function is introduced), and averages noise down when the
	       output pixels are larger than the input pixels. However, it
	       can suffer from ringing around very bright features, and an
	       all-positive input map may end up with small negative values.
	       "spline" instead uses spline interpolation to look up the
	       value in the intput map corresponding to each pixel center
	       in the output map. The spline order is controlled with the
	       "order" argument. Overall "harm" is best suited for normal
	       sky maps, while "spline" with order = 0 or 1 is best suited
	       for hitcount maps and masks.
	order: The spline order to use when method="spline".
	       0 corresponds to nearest neighbor interpolation.
	       1 corresponds to bilinear interpolation (default)
	       Higher order interpolation is not supported - use
	       method="harm" for that.
	extensive: Whether the map represents an extensive (as opposed to
	       intensive) quantity. Extensive quantities have values
	       proportional to the pixel size, unlike intensive quantities.
	       Hitcount per pixel is an extensive quantity. Hitcount per
	       square degree is an intensive quantity, as is a temperature
	       map. Defaults to False.
	bsize: The spline method operates on batches of pixels to save memory.
	       This controls the batch size, in pixels. Defaults to 100000.
	verbose: Whether to print information about what it's doing.
	       Defaults to False, which doesn't print anything.

	Typical usage:
	* map  = healpix2map(map_healpix,  shape, wcs, rot="gal,cel")
	* ivar = healpix2map(ivar_healpix, shape, wcs, rot="gal,cel", method="spline", spin=[0], extensive=True)
	"""
	iheal    = np.asarray(iheal)
	npix     = iheal.shape[-1]
	nside    = curvedsky.npix2nside(npix)
	ipixsize = 4*np.pi/npix
	if out is None:
		out = enmap.zeros(iheal.shape[:-1]+shape[-2:], wcs, dtype=iheal.dtype)
	else: shape, wcs = out.geometry
	if lmax is None: lmax = 3*nside
	if method in ["harm", "harmonic"]:
		# Harmonic interpolation preserves the power spectrum, but can introduce ringing.
		# Probably not a good choice for positive-only quantities like hitcounts.
		# Coordinate rotation is slow.
		alm = curvedsky.map2alm_healpix(iheal, lmax=lmax, spin=spin, niter=niter)
		if rot is not None:
			curvedsky.rotate_alm(alm, *rot2euler(rot), inplace=True)
		curvedsky.alm2map(alm, out, spin=spin)
		del alm
	elif method == "spline":
		# Covers linear interpolation (order=1) and nearest neighbor (order=0).
		# Coordinate rotation may be slow.
		import healpy
		if order > 1:
			raise ValueError("Only order 0 and order 1 spline interpolation supported from healpix maps")
		# Figure out if we need to compute polarization rotations
		pol  = iheal.ndim > 1 and any([s != 0 for s,c1,c2 in enmap.spin_helper(spin, iheal.shape[-2])])
		# Batch to save memory
		brow = (bsize+out.shape[-1]-1)//out.shape[-1]
		for i1 in range(0, out.shape[-2], brow):
			i2   = min(i1+brow, out.shape[-2])
			pos  = out[...,i1:i2,:].posmap().reshape(2,-1)[::-1]
			if rot is not None:
				# Not sure why the [::-1] is necessary here. Maybe psi,theta,phi vs. phi,theta,psi?
				pos = coordinates.transform_euler(inv_euler(rot2euler(rot))[::-1], pos, pol=pol)
			pos[1] = np.pi/2 - pos[1]
			if order == 0:
				# Nearest neighbor. Just read off from the pixels
				vals = iheal[...,healpy.ang2pix(nside, pos[1], pos[0])]
			else:
				# Bilinear interpolation. healpy only supports one component at a time, so loop
				vals = np.zeros(iheal.shape[:-1]+pos.shape[-1:], iheal.dtype)
				for I in utils.nditer(iheal.shape[:-1]):
					vals[I] = healpy.get_interp_val(iheal[I], pos[1], pos[0])
			if rot is not None and iheal.ndim > 1:
				# Update the polarization to account for the new coordinate system
				for s, c1, c2 in enmap.spin_helper(spin, iheal.shape[-2]):
					vals = enmap.rotate_pol(vals, -pos[2], spin=s, comps=[c1,c2-1], axis=-2)
			out[...,i1:i2,:] = vals.reshape(vals.shape[:-1]+(i2-i1,-1))
	else:
		raise ValueError("Map reprojection method '%s' not recognized" % str(method))
	if extensive:
		out *= out.pixsizemap(broadcastable=True)/ipixsize
	return out

def rot2euler(rot):
	"""Given a coordinate rotation description, return the [rotz,roty,rotz] euler
	angles it corresponds to. The rotation desciption can either be those angles
	directly, or a string of the form isys,osys"""
	gal2cel = np.array([57.06793215,  62.87115487, -167.14056929])*utils.degree
	if isinstance(rot, basestring):
		try: isys, osys = rot.split(",")
		except ValueError:
			raise ValueError("Rotation string must be of form 'isys,osys', but got '%s'" % str(rot))
		R = spatial.transform.Rotation.identity()
		# Handle input system
		if   isys in ["cel","equ"]: pass
		elif isys == "gal": R *= spatial.transform.Rotation.from_euler("zyz", gal2cel)
		else: raise ValueError("Unrecognized system '%s'" % isys)
		# Handle output system
		if   osys in ["cel","equ"]: pass
		elif osys == "gal": R *= spatial.transform.Rotation.from_euler("zyz", gal2cel).inv()
		else: raise ValueError("Unrecognized system '%s'" % osys)
		return R.as_euler("zyz")
	else:
		rot = utils.asfarray(rot)
		return rot

def inv_euler(euler): return [-euler[2], -euler[1], -euler[0]]

def restrict_nside(nside, mode="mul32", round="ceil"):
	"""Given an arbitrary Healpix nside, return one that's restricted in
	various ways according to the "mode" argument:

	"pow2":  Restrict to a power of 2. This is required for compatibility
	 with the rarely used "nest" pixel ordering in Healpix, and is the standard
	 in the Healpix world.
	"mul32": Restrict to multiple of 32, unless 12*nside**2<=1024.
	 This is enough to make the maps writable by healpy.
	"any":   No restriction

	The "round" argument controls how any rounding is done. This can be one
	of the strings "ceil" (default), "round" or "floor", or you can pass in
	a custom function(nside) -> nside.

	In all cases, the final nside is converted to an integer and capped to
	1 below.
	"""
	if isinstance(round, basestring):
		round = {"floor":np.floor, "round":np.round, "ceil":np.ceil}[round]
	if   mode == "any": nside = round(nside)
	elif mode == "mul32":
		if 12*nside**2 > 1024:
			nside = round(nside/32)*32
	elif mode == "pow2":
		nside = 2**round(np.log2(nside))
	else:
		raise ValueError("Unrecognized nside mode '%s'" % str(mode))
	nside = max(1,int(nside))
	return nside


################################
####### Old stuff below ########
################################

def centered_map(imap, res, box=None, pixbox=None, proj='car', rpix=None,
				 width=None, height=None, width_multiplier=1.,
				 rotate_pol=True, **kwargs):
	raise NotImplementedError("This function has been removed.")
	
def healpix_from_enmap_interp(imap, **kwargs):
	raise RuntimeError("This function has been removed. Use reproject.map2healpix(...method='spline').")


def healpix_from_enmap(imap, lmax, nside):
	raise RuntimeError("This function has been removed. Use reproject.map2healpix(...method='harm').")



def enmap_from_healpix(hp_map, shape, wcs, ncomp=1, unit=1, lmax=0,
					   rot="gal,equ", first=0, is_alm=False, return_alm=False, f_ell=None):
	raise RuntimeError("This function has been removed. Use reproject.healpix2map(...method='harm').")


def enmap_from_healpix_interp(hp_map, shape, wcs , rot="gal,equ",
							  interpolate=False):
	raise RuntimeError("This function has been removed. Use reproject.healpix2map(...method='spline').")



def ivar_hp_to_cyl(hmap, shape, wcs, rot=False,do_mask=True,extensive=True):
	raise NotImplementedError("This function has been removed.")

# Helper functions


def gnomonic_pole_wcs(shape, res):
	raise NotImplementedError("This function has been removed.")
 
def gnomonic_pole_geometry(width, res, height=None):
	raise NotImplementedError("This function has been removed.")


def rotate_map(imap, shape_target=None, wcs_target=None, shape_source=None,
			   wcs_source=None, pix_target=None, **kwargs):
	raise NotImplementedError("This function has been removed.")


def get_rotated_pixels(shape_source, wcs_source, shape_target, wcs_target,
					   inverse=False, pos_target=None,
					   center_target=None, center_source=None):
	raise NotImplementedError("This function has been removed.")

def cutout(imap, width=None, ra=None, dec=None, pad=1, corner=False,
		   res=None, npix=None, return_slice=False,sindex=None):
	raise NotImplementedError("This function has been removed.")


def rect_box(width, center=(0., 0.), height=None):
	raise NotImplementedError("This function has been removed.")

def get_pixsize_rect(shape, wcs):
	raise NotImplementedError("This function has been removed.")

def rect_geometry(width, res, height=None, center=(0., 0.), proj="car"):
	raise NotImplementedError("This function has been removed.")


def distribute(N,nmax):
	"""
	Distribute N things into cells as equally as possible such that 
	no cell has more than nmax things.
	"""
	actual_max = int(2.*(nmax+1)/3.)
	numcells = int(round(N*1./actual_max))
	each_cell = [actual_max]*(numcells-1)
	rem = N-sum(each_cell)
	if rem>0: each_cell.append(rem)
	assert sum(each_cell)==N
	return each_cell

def populate(shape,wcs,ofunc,maxpixy = 400,maxpixx = 400):
	"""
	Loop through tiles in a new map of geometry (shape,wcs)
	with tiles that have maximum allowed shape (maxpixy,maxpixx)
	such that each tile is populated with the result of
	ofunc(oshape,owcs) where oshape,owcs is the geometry of each
	tile.
	"""
	omap = enmap.zeros(shape,wcs)
	Ny,Nx = shape[-2:]
	tNys = distribute(Ny,maxpixy)
	tNxs = distribute(Nx,maxpixx)
	numy = len(tNys)
	numx = len(tNxs)
	sny = 0
	ntiles = numy*numx
	print("Number of tiles = ",ntiles)
	done = 0
	for i in range(numy):
		eny = sny+tNys[i]
		snx = 0
		for j in range(len(tNxs)):
			enx = snx+tNxs[j]
			sel = np.s_[...,sny:eny,snx:enx]
			oshape,owcs = enmap.slice_geometry(shape,wcs,sel)
			omap[sel] = ofunc(oshape,owcs)
			snx += tNxs[j]
			done += 1
		sny += tNys[i]
		print(done , " / ", ntiles, " tiles done...")
	return omap


def postage_stamp(inmap, ra_deg, dec_deg, width_arcmin,
		  res_arcmin, proj='gnomonic', return_cutout=False,
		  npad=3, rotate_pol=True, **kwargs):
	raise RuntimeError("postage_stamp has been removed. Please use thumbnails instead.")




"""This module handles resampling of time-series and similar arrays."""
import numpy as np
from . import utils, fft

def resample(d, factors=[0.5], axes=None, method="fft"):
	factors = np.atleast_1d(factors)
	if np.allclose(factors,1): return d
	if method == "fft":
		if axes is None: axes = range(-len(factors),0)
		lens = [int(d.shape[ax]*fact+0.5) for ax, fact in zip(axes, factors)]
		return resample_fft(d, lens, axes)
	elif method == "bin":
		return resample_bin(d, factors, axes)
	else:
		raise NotImplementedError("Resampling method '%s' is not implemented" % method)

def resample_bin(d, factors=[0.5], axes=None):
	if np.allclose(factors,1): return d
	down = [max(1,int(round(1/f))) for f in factors]
	up   = [max(1,int(round(f)))   for f in factors]
	d    = downsample_bin(d, down, axes)
	return upsample_bin  (d, up, axes)

def downsample_bin(d, steps=[2], axes=None):
	assert len(steps) <= d.ndim
	if axes is None: axes =range(-len(steps),0)
	assert len(axes) == len(steps)
	# Expand steps to cover every axis in order
	fullsteps = np.zeros(d.ndim,dtype=int)+1
	for ax, step in zip(axes, steps): fullsteps[ax]=step
	# Make each axis an even number of steps to prepare for reshape
	s = tuple([slice(0,L//step*step) for L,step in zip(d.shape,fullsteps)])
	d = d[s]
	# Reshape each axis to L/step,step to prepare for mean
	newshape = np.concatenate([[L//step,step] for L,step in zip(d.shape,fullsteps)])
	d = np.reshape(d, newshape)
	# And finally take the mean over all the extra axes
	return np.mean(d, tuple(range(1,d.ndim,2)))

def upsample_bin(d, steps=[2], axes=None):
	shape = d.shape
	assert len(steps) <= d.ndim
	if axes is None: axes = np.arange(-1,-len(steps)-1,-1)
	assert len(axes) == len(steps)
	# Expand steps to cover every axis in order
	fullsteps = np.zeros(d.ndim,dtype=int)+1
	for ax, step in zip(axes, steps): fullsteps[ax]=step
	# Reshape each axis to (n,1) to prepare for tiling
	newshape = np.concatenate([[L,1] for L in shape])
	d = np.reshape(d, newshape)
	# And tile
	d = np.tile(d, np.concatenate([[1,s] for s in fullsteps]))
	# Finally reshape back to proper dimensionality
	return np.reshape(d, np.array(shape)*np.array(fullsteps))

def resample_fft(d, n, axes=None):
	"""Resample numpy array d via fourier-reshaping. Requires periodic data.
	n indicates the desired output lengths of the axes that are to be
	resampled. By default the last len(n) axes are resampled, but this
	can be controlled via the axes argument."""
	d = np.asanyarray(d)
	# Compute output lengths from factors if necessary
	n = np.atleast_1d(n)
	if axes is None: axes = np.arange(-len(n),0)
	else: axes = np.atleast_1d(axes)
	if len(n) == 1: n = np.repeat(n, len(axes))
	else: assert len(n) == len(axes)
	assert len(n) <= d.ndim
	# Nothing to do?
	if np.all(d.shape[-len(n):] == n): return d
	# Use the simple version if we can. It has lower memory overhead
	if d.ndim == 2 and len(n) == 1 and (axes[0] == 1 or axes[0] == -1):
		return resample_fft_simple(d, n[0])
	# Perform the fourier transform
	fd = fft.fft(d, axes=axes)
	# Frequencies are 0 1 2 ... N/2 (-N)/2 (-N)/2+1 .. -1
	# Ex 0* 1 2* -1 for n=4 and 0* 1 2 -2 -1 for n=5
	# To upgrade,   insert (n_new-n_old) zeros after n_old/2
	# To downgrade, remove (n_old-n_new) values after n_new/2
	# The idea is simple, but arbitrary dimensionality makes it
	# complicated.
	norm = 1.0
	for ax, nnew in zip(axes, n):
		ax %= d.ndim
		nold = d.shape[ax]
		dn   = nnew-nold
		if dn > 0:
			padvals = np.zeros(fd.shape[:ax]+(dn,)+fd.shape[ax+1:],fd.dtype)
			spre  = tuple([slice(None)]*ax+[slice(0,nold//2)]+[slice(None)]*(fd.ndim-ax-1))
			spost = tuple([slice(None)]*ax+[slice(nold//2,None)]+[slice(None)]*(fd.ndim-ax-1))
			fd = np.concatenate([fd[spre],padvals,fd[spost]],axis=ax)
		elif dn < 0:
			spre  = tuple([slice(None)]*ax+[slice(0,nnew//2)]+[slice(None)]*(fd.ndim-ax-1))
			spost = tuple([slice(None)]*ax+[slice(nnew//2-dn,None)]+[slice(None)]*(fd.ndim-ax-1))
			fd = np.concatenate([fd[spre],fd[spost]],axis=ax)
		norm *= float(nnew)/nold
	# And transform back
	res  = fft.ifft(fd, axes=axes, normalize=True)
	del fd
	res *= norm
	return res if np.issubdtype(d.dtype, np.complexfloating) else res.real

def resample_fft_simple(d, n, ngroup=100):
	"""Resample 2d numpy array d via fourier-reshaping along
	last axis."""
	nold = d.shape[1]
	if n == nold: return d
	res  = np.zeros([d.shape[0],n],dtype=d.dtype)
	dn   = n-nold
	for di in range(0, d.shape[0], ngroup):
		fd = fft.fft(d[di:di+ngroup])
		if n < nold:
			fd = np.concatenate([fd[:,:n//2],fd[:,n//2-dn:]],1)
		else:
			fd = np.concatenate([fd[:,:nold//2],np.zeros([len(fd),n-nold],fd.dtype),fd[:,nold//2:]],-1)
		res[di:di+ngroup] = fft.ifft(fd, normalize=True).real
		del fd
	res *= float(n)/nold
	return res

def make_equispaced(d, t, quantile=0.1, order=3, mask_nan=False):
	"""Given an array d[...,nt] of data that has been sampled at times t[nt],
	return an array that has been resampled to have a constant sampling rate."""
	# Find the typical sampling rate of the input. We will lose information if
	# we don't use a sampling rate that's higher than the highest rate in the
	# input. But we also don't want to exaggerate the number of samples. Use a
	# configurable quantile as a compromise.
	dt   = np.percentile(np.abs(t[1:]-t[:-1]), quantile*100)
	# Modify so we get a whole number of samples
	nout = utils.nint(np.abs(t[-1]-t[0])/dt)+1
	dt   = (t[-1]-t[0])/(nout-1)
	# Construct our output time steps
	tout = np.arange(nout)*dt + t[0]
	# To interpolate, we need the input sample number as a function of time
	samples = np.interp(tout, t, np.arange(len(t)))
	# Now that we have the samples we can finally evaluate the function
	dout = utils.interpol(d, samples[None], mode="nearest", order=order, mask_nan=mask_nan)
	return dout, tout



import cython, numpy as np
cimport numpy as np
from libc.stdlib cimport free, calloc
# I had to call this distances_wrap instead of just distances because
# .pdx files with the same name get automatically imported into the current
# namespace, which means I would have to give stupid names to the functions here
cimport srcsim_wrap as c

__version__ = 1.0

def sim_objects(float[:,:,::1] map, float[::1] obj_decs, float[::1] obj_ras, int[::1] obj_ys, int[::1] obj_xs, float[:,::1] amps, profs, int[::1] prof_ids, posmap, float vmin, float rmax=0, separable=False, transpose=False, prof_equi=False, op="add", int csize=8, return_times=False):
	"""
	map: array[ncomp,ny,nx]. Caller should make sure it has exactly 3 dims
	poss:   [{ra,dec},nobj] float
	pixs:   [{x,y},nobj] float
	amps:   [ncomp,nobj] float
	profs:  [nprof][{r,val}] float, can have different length for each profile
	prof_ids: [nprof] int
	posmap: [dec[ny],ra[nx]] if separable, otherwise [{dec,ra}][ny,nx]
	vmin:   The lowest value to evaluate profiles to. If this is set too low combined
	        with profiles that never reach zero, then things will get very slow.

	op:     How to combine contributions from multiple simulated objects and the input map.
	        add: Things add together linearly
	        max: Each pix will be the max of all contributions to that pixel
	        min: Each pix will be the min of all contributions to that pixel
	csize:   Size of cells used internally when determining which parts of the sky to
	         consider for each source, in pixels.
	
	Returns the resulting map. If inplace, then this will be the same object as the
	map that was passed in, otherwise it will be a new map.
	"""
	# Prepare our inputs. First the map
	cdef int ncomp = map.shape[0]
	if ncomp == 0: return map # nothing to do
	cdef int ny = map.shape[1]
	cdef int nx = map.shape[2]
	# I can't figure out how to make a memoryview of an array of pointers, so let's just do
	# things brute force
	cdef float ** map_ = <float**>calloc(ncomp, sizeof(float*));
	for i in range(ncomp):
		map_[i] = &map[i,0,0]
	# Then the positions
	cdef int nobj = len(obj_ras)
	assert len(obj_decs) == len(obj_xs) == len(obj_ys) == nobj
	# and amplitudes
	assert amps.shape[0] == ncomp and amps.shape[1] == nobj, "amps [%d,%d] must be [ncomp=%d,nobj=%d]" % (amps.shape[0], amps.shape[1], ncomp, nobj)
	cdef float ** amps_ = <float**>calloc(ncomp, sizeof(float*))
	for i in range(ncomp):
		amps_[i] = &amps[i,0]
	# The profiles are variable-length
	cdef int nprof = len(profs)
	cdef int[::1] prof_ns = np.zeros(nprof,dtype=np.int32)
	cdef float ** prof_rs = <float**>calloc(nprof, sizeof(float*))
	cdef float ** prof_vs = <float**>calloc(nprof, sizeof(float*))
	cdef float[:,::1] prof
	for i in range(nprof):
		prof = np.asanyarray(profs[i], dtype=np.float32, order="C")
		assert prof.ndim == 2 and len(prof) == 2, ("prof must be [{r,val},:], but #%d is not" % i)
		prof_ns[i] = prof.shape[1]
		prof_rs[i] = &prof[0,0]
		prof_vs[i] = &prof[1,0]
	cdef int prof_equi_ = prof_equi
	# Profile ids
	assert len(prof_ids) == nobj, "prof_ids (%d) must be [nobj=%d]" % (len(prof_ids), nobj)
	# pixel positions
	cdef int separable_ = separable
	assert len(posmap) == 2, "posmap must be [dec[ny],ra[nx]] if separable, else [{dec,ra},ny,nx]"
	pix_decs = np.asanyarray(posmap[0], dtype=np.float32, order="C")
	pix_ras  = np.asanyarray(posmap[1], dtype=np.float32, order="C")
	if separable:
		assert pix_decs.shape == (ny,) and pix_ras.shape == (nx,), "posmap ([%d],[%d]) must be [dec[ny=%d],ra[nx=%d]] if separable" % (len(pix_decs), len(pix_ras), ny, nx)
	else:
		assert pix_decs.shape == (ny,nx) and pix_ras.shape == (ny,nx), "posmap must be [{dec,ra},ny,nx] if not separable"
	cdef float [::1] pix_decs_ = pix_decs.reshape(-1)
	cdef float [::1] pix_ras_  = pix_ras.reshape(-1)
	# The rest are simple
	cdef int op_ = {"add":0, "max":1, "min":2}[op]
	times = np.zeros(3)
	cdef double [::1] times_ = times
	# Phew! That's a lot of wrapping!
	c.sim_objects(nobj, &obj_decs[0], &obj_ras[0], &obj_ys[0], &obj_xs[0], &amps_[0], nprof, &prof_ids[0], &prof_ns[0], &prof_rs[0], &prof_vs[0], prof_equi, vmin, rmax, op_, ncomp, ny, nx, separable, transpose, &pix_decs_[0], &pix_ras_[0], &map_[0], &map_[0], csize, &times_[0])
	free(map_)
	free(amps_)
	free(prof_rs)
	free(prof_vs)
	if return_times: return map, times
	else:            return map

def radial_sum(float[:,:,::1] map, float[::1] obj_decs, float[::1] obj_ras, int[::1] obj_ys, int[::1] obj_xs, float[::1] rs, posmap, profs=None, separable=False, prof_equi=False, return_times=False):
	"""
	map: array[ncomp,ny,nx]. Caller should make sure it has exactly 3 dims
	poss:   [{ra,dec},nobj] float
	pixs:   [{x,y},nobj] float
	rs:     [nbin+1] float, bin edges (ascending). Faster if equi-spaced starting at 0
	posmap: [dec[ny],ra[nx]] if separable, otherwise [{dec,ra}][ny,nx]

	Returns the resulting profiles. If inplace, then this will be the same object as the
	map that was passed in, otherwise it will be a new map.
	"""
	# Prepare our inputs. First the map
	cdef int ncomp = map.shape[0]
	if ncomp == 0: return map # nothing to do
	cdef int ny = map.shape[1]
	cdef int nx = map.shape[2]
	# I can't figure out how to make a memoryview of an array of pointers, so let's just do
	# things brute force
	cdef float ** map_ = <float**>calloc(ncomp, sizeof(float*));
	for i in range(ncomp):
		map_[i] = &map[i,0,0]
	# Then the positions
	cdef int nobj = len(obj_ras)
	assert len(obj_decs) == len(obj_xs) == len(obj_ys) == nobj
	# Bin definitions
	cdef int nbin = len(rs)-1
	assert nbin >= 2, "rs must be [nbin] where nbin >= 2"
	# Output profiles
	if profs is None: profs = np.zeros((nobj,ncomp,nbin),np.float32)
	cdef float[:,:,::1] profs_buf = profs
	cdef float *** profs_ = <float***>calloc(nobj, sizeof(float**));
	for oi in range(nobj):
		profs_[oi] = <float**>calloc(ncomp, sizeof(float*));
		for ci in range(ncomp):
			profs_[oi][ci] = &profs_buf[oi,ci,0]
	# pixel positions
	cdef int separable_ = separable
	cdef int prof_equi_ = prof_equi
	assert len(posmap) == 2, "posmap must be [dec[ny],ra[nx]] if separable, else [{dec,ra},ny,nx]"
	pix_decs = np.asanyarray(posmap[0], dtype=np.float32, order="C")
	pix_ras  = np.asanyarray(posmap[1], dtype=np.float32, order="C")
	if separable:
		assert pix_decs.shape == (ny,) and pix_ras.shape == (nx,), "posmap ([%d],[%d]) must be [dec[ny=%d],ra[nx=%d]] if separable" % (len(pix_decs), len(pix_ras), ny, nx)
	else:
		assert pix_decs.shape == (ny,nx) and pix_ras.shape == (ny,nx), "posmap must be [{dec,ra},ny,nx] if not separable"
	cdef float [::1] pix_decs_ = pix_decs.reshape(-1)
	cdef float [::1] pix_ras_  = pix_ras.reshape(-1)
	times = np.zeros(1)
	cdef double [::1] times_ = times
	# Phew! That's a lot of wrapping!
	c.radial_sum(nobj, &obj_decs[0], &obj_ras[0], &obj_ys[0], &obj_xs[0], nbin, &rs[0], prof_equi_, ncomp, ny, nx, separable_, &pix_decs_[0], &pix_ras_[0], map_, profs_, &times_[0])
	free(map_)
	for oi in range(nobj): free(profs_[oi])
	free(profs_)
	if return_times: return profs, times
	else:            return profs



import numpy as np, numba, ducc0, time
from . import coordinates, enmap, utils, curvedsky, wcsutils

beta    = 0.001235
dir_equ = np.array([167.919, -6.936])*np.pi/180
dir_gal = np.array([263.986, 48.247])*np.pi/180
dir_ecl = np.array([171.640,-11.154])*np.pi/180
from .utils import T_cmb, h, c, k

def boost_map(map, dir=dir_equ, beta=beta, modulation="T2lin", T0=utils.T_cmb, freq=150e9,
		return_modulation=False, dipole=False, map_unit=1e-6, spin=[0,2], aberrate=True,
		modulate=True, nthread=None, coord_dtype=None, boundary="auto"):
	"""Doppler-boost (aberrate and modulate) the given input map map. The doppler boost
	goes in direction dir[{ra,dec}] with speed beta in units of c.

	spin controls how the 3rd last axis is interpreted. spin = [0,2] means that the
	first entry is scalar and the next pair is spin-2. So this works for a [TQU,ny,nx]
	map. It is safe to pass [0,2] even when the map is just scalar.

	The modulation argument covers five cases. The default is "T2lin".
	* None: No modulation or unit conversion is performed at all.
	* "plain" or "T2T": The input and output maps are in real temperature units.
	* "T2lin": The input map is real temperature, but the output map is in
	  linearized temperature units, which are proportional to the flux density.
	  This is appropriate for simulating observed CMB maps.
	* "lin2T": The opposite of T2lin. Appropriate for deaberration.
	* "lin2lin": Both input and output maps are in linearized units. Probably
	  not very useful.

	The boundary conditions are always periodic along the x axis. For the y axis
	they are controlled by the boundary argument, which has the following values:
	* "periodic": map[Ny,x] = map[0,x]
	* "fullsky":  map[Ny,x] = map[Ny-1,x+Nx/2]. This is slower and uses more memory
	  than periodic.
	* "auto": Choose fullsky if map covers close to the full sky in the y direction,
	  otherwise periodic. This is the default.

	If dipole is True, then monopole-related terms will be included in the modulation.
	This is mainly the CMB dipole.

	If return_modulation == True, then a tuple of (omap, A) will be returned, where
	A is the modulation as returned from calc_boost. This can be useful if one wants
	to compute the same map modulated at several different frequencies.
	
  The modulation and aberration steps can be individually skipped using the
  corresponding arguments. By default both are performed.
	"""
	if return_modulation: assert modulate, "Can't return modulation of modulation is disabled"
	if aberrate:
		map    = aberrate_map(map, dir=dir, beta=beta, spin=spin, nthread=nthread, boundary=boundary)
	if modulate:
		map, A = modulate_map(map, dir=dir, beta=beta, spin=spin, nthread=nthread, dipole=dipole,
			modulation=modulation, T0=T0, freq=freq, map_unit=map_unit, return_modulation=True)
	if return_modulation: return map, A
	else: return map

def deboost_map(map, dir=dir_equ, beta=beta, modulation="lin2T", T0=utils.T_cmb, freq=150e9,
		return_modulation=False, dipole=False, map_unit=1e-6, spin=[0,2], aberrate=True,
		modulate=True, nthread=None, coord_dtype=None, boundary="auto"):
	"""The inverse of boost_map. Simply calls boost map with the sign of beta flipped, and
	with a default modulation of lin2T instead of T2lin."""
	return boost_map(map, dir=dir, beta=-beta, modulation=modulation, T0=T0, freq=freq,
		return_modulation=return_modulation, dipole=dipole, map_unit=map_unit, spin=spin,
		aberrate=aberrate, modulate=modulate, nthread=nthread, coord_dtype=coord_dtype,
		boundary=boundary)

def aberrate_map(map, dir=dir_equ, beta=beta, spin=[0,2],
		nthread=None, coord_dtype=None, boundary="auto"):
	"""Perform the aberration part of the doppler boost. See
	boost_map for details."""
	if coord_dtype is None: coord_dtype = map.dtype
	return Aberrator(map.shape, map.wcs, dir=dir, beta=beta, spin=spin,
		nthread=nthread, coord_dtype=coord_dtype, boundary=boundary)(map)

def deaberrate_map(map, dir=dir_equ, beta=beta, spin=[0,2],
		nthread=None, coord_dtype=None, boundary="auto"):
	return aberrate_map(map, dir=dir, beta=-beta, spin=spin,
		nthread=nthread, coord_dtype=coord_dtype, boundary=boundary)

def modulate_map(map, dir=dir_equ, beta=beta,
		modulation="T2lin", T0=utils.T_cmb, freq=150e9, return_modulation=False,
		dipole=False, map_unit=1e-6, spin=[0,2], nthread=None):
	"""Perform the modulation part of the doppler boost. See boost_map
	for details."""
	modulator = Modulator(map.shape, map.wcs, dir=dir, beta=beta, spin=spin,
		modulation=modulation, T0=T0, freq=freq, dipole=dipole, map_unit=map_unit,
		nthread=nthread, dtype=map.dtype)
	map = modulator(map)
	if return_modulation: return map, modulator.A
	else: return map

def demodulate_map(map, dir=dir_equ, beta=beta,
		modulation="lin2T", T0=utils.T_cmb, freq=150e9, return_modulation=False,
		dipole=False, map_unit=1e-6, spin=[0,2], nthread=None):
	return modulate_map(map, dir=dir, beta=-beta, modulation=modulation,
		T0=T0, freq=freq, return_modulation=return_modulation, dipole=dipole,
		map_unit=map_unit, spin=spin, nthread=nthread)

# Classes for repeat calcuations with similar maps

class Aberrator:
	def __init__(self, shape, wcs, dir=dir_equ, beta=beta, spin=[0,2],
			nthread=None, coord_dtype=np.float64, boundary="auto"):
		"""Construct an Aberrator object, that can be used to more efficiently
		aberrate a map. E.g.

		  aberrator = Aberrator(shape, wcs)
		  for map in maps:
		    omap = aberrator(map)
		    do something with omap

		This should be faster than repeatedly calling aberrate_map because
		it precomputes the deflection information instead of needing to
		compute it from scratch each time."""
		coord_ctype = utils.complex_dtype(coord_dtype)
		nthread = int(utils.fallback(utils.getenv("OMP_NUM_THREADS",nthread),0))
		# 1. Calculate the aberration field. These are tiny
		alm_dpos = calc_boost_field(-beta, dir, nthread=nthread)
		# 2. Evaluate these on our target geometry.
		deflect = enmap.zeros(alm_dpos.shape[:-1]+shape[-2:], wcs, coord_dtype)
		curvedsky.alm2map(alm_dpos.astype(coord_ctype, copy=False), deflect, spin=1, nthread=nthread)
		# 3. Calculate the offset angles.
		rinfo = curvedsky.get_ring_info(shape, wcs)
		dphi  = np.full(shape[-2], wcs.wcs.cdelt[0]*utils.degree)
		odec, ora, gamma = ducc0.misc.get_deflected_angles(theta=rinfo.theta, phi0=rinfo.phi0,
			nphi=rinfo.nphi, dphi=dphi, ringstart=rinfo.offsets, nthreads=nthread, calc_rotation=True,
			deflect=np.asarray(deflect).reshape(2,-1).T).T
		del deflect
		odec = np.pi/2-odec
		gamma= enmap.ndmap(gamma.reshape(shape[-2:]), wcs)
		# 4. Calculate pixel coordinates of offset angles. In general this would use
		# enmap.sky2pix, but that's slow. Much faster for our typical projections.
		# Probably worth it to make overrides.
		pix = sky2pix(shape, wcs, [odec, ora]).astype(coord_dtype, copy=False)
		# Make the x pixel 0:nphi, assuming it's at most one period away
		fast_rewind(pix[1].reshape(-1), rinfo.nphi[0])
		del odec, ora
		# 5. Evaluate the map at these locations using nufft
		# Boundary condition
		if boundary == "auto": boundary = "fullsky" if fully(shape, wcs) else "periodic"
		if boundary not in ["periodic", "fullsky"]:
			raise ValueError("Unrecognized boundary '%s'" % str(boundary))
		# 6. Store for when the map is passed in later
		self.pix       = pix
		self.gamma     = gamma
		self.nthread   = nthread
		self.spin      = spin
		self.boundary  = boundary
	def __call__(self, map, spin=None, nthread=None):
		if nthread is None: nthread = self.nthread
		if spin    is None: spin    = self.spin
		shape, wcs = map.shape, map.wcs
		ydouble    = self.boundary == "fullsky"
		map = interpol_map(map, self.pix, nthread=nthread, ydouble=ydouble)
		map = enmap.ndmap (map.reshape(shape), wcs)
		# 6. Apply polarization rotation. ducc0.misc.lensing_rotate can do this,
		# but for some reason it operates on complex numbers instead of a QU field.
		# So seems like I would have to waste time and memory transforming from/to
		# this ordering. We loop over pre-dimensions to reduce memory use
		for s, I in enmap.spin_pre_helper(spin, map.shape[:-2]):
			rotate_pol(utils.fix_zero_strides(map[I]), -self.gamma, spin=s)
		return map

class Modulator:
	def __init__(self, shape, wcs, dir=dir_equ, beta=beta,
			modulation="T2lin", T0=utils.T_cmb, freq=150e9, dipole=False,
			map_unit=1e-6, spin=[0,2], dtype=np.float64, nthread=None):
		"""Construct a Modulator object, that can be used to more efficiently
		modulate a map. E.g.

		  modulator = Modulator(shape, wcs)
		  for map in maps:
		    omap = modulator(map)
		    do something with omap

		This should be faster than repeatedly calling modulate_map because
		it precomputes the modulation information instead of needing to
		compute it from scratch each time."""
		nthread = int(utils.fallback(utils.getenv("OMP_NUM_THREADS",nthread),0))
		# 1. Calculate the aberration field. These are tiny
		alm_mod = calc_boost_field(-beta, dir, nthread=nthread, modulation=True, mod_exp=-1)[1]
		alm_mod = alm_mod.astype(utils.complex_dtype(dtype), copy=False)
		# 2: Apply modulation
		A = enmap.zeros(alm_mod.shape[:-1]+shape[-2:], wcs, dtype)
		curvedsky.alm2map(alm_mod, A, spin=0, nthread=nthread)
		# Store for __call__
		self.nthread = nthread;  self.A     = A;     self.modulation = modulation
		self.T0      = T0;       self.freq  = freq;  self.dipole     = dipole
		self.dtype   = dtype;    self.spin  = spin;  self.map_unit   = map_unit
	def __call__(self, map, spin=None):
		if spin is None: spin = self.spin
		if map.dtype != self.dtype:
			warnings.warn("Modulator dtype does not match argument dtype")
		return apply_modulation(map, self.A, spin=spin, T0=self.T0, freq=self.freq,
			map_unit=self.map_unit, mode=self.modulation, dipole=self.dipole)

def calc_boost_1d(z, beta):
	"""Given the z, the cosine of the angle from the direction of travel in
	the CMB rest frame, and beta, the speed as a fraction of c, compute the
	observed z_obs after taking into account the effect of aberration, as well
	as the corresponding modulation A, such that T_obs(z_obs) = A * T_rest(z).
	Pass in -beta to get the reverse transformation, from obs to rest."""
	gamma = (1-beta**2)**-0.5
	z_obs = (z+beta)/(1+z*beta)
	np.clip(z_obs, -1, 1, out=z_obs)
	A     = 1/(gamma*(1-z_obs*beta))
	return z_obs, A

#### Helpers ####

def fully(shape, wcs, tol=0.1):
	minfo = curvedsky.analyse_geometry(shape, wcs)
	if minfo.ducc_geo is None: return False
	yfrac  = abs(shape[-2]/minfo.ducc_geo.ny)
	return yfrac > 1-tol

def beta2lmax(beta):
	# Infer lmax from beta. Empirical formula, valid to high boost factors.
	beta  = np.abs(beta)
	gamma = (1-beta**2)**-0.5
	lmax  = utils.ceil(1/(4e-3+1/gamma)**0.62*14+3.5)
	return lmax

def calc_boost_field(beta, dir, lmax=None, nthread=None, modulation=False, mod_exp=1):
	if lmax is None:
		lmax=beta2lmax(beta)
	nthread = int(utils.fallback(utils.getenv("OMP_NUM_THREADS",nthread),0))
	n      = lmax+2
	# zenith angles to sample the sky at, in fejer1
	itheta = np.arange(0,n)*np.pi/(n-1)
	iz     = np.cos(itheta)
	oz,  A = calc_boost_1d(iz, beta)
	otheta = np.arccos(oz)
	# Calculate the local position deflection. We have
	# no deflection in phi in this coordinate system
	dpos   = np.zeros([2,n])
	dpos[0]= otheta-itheta
	alm    = curvedsky.prof2alm(dpos, dir=dir, spin=1, nthread=nthread)
	if modulation:
		A  **= mod_exp
		malm = curvedsky.prof2alm(A, dir=dir, spin=0, nthread=nthread)
		return alm, malm
	else:
		return alm

def interpol_map(imap, pixs, epsilon=None, nthread=None, ydouble=False):
	ny, nx = imap.shape[-2:]
	if ydouble:
		# Make y doubled map. This is necessary for the correct boundary condition for
		# fullsky maps, but makes less sense for partial sky maps
		dmap = enmap.zeros(imap.shape[:-2]+(2*ny,nx), imap.wcs, imap.dtype)
		dmap[...,:ny,:] = imap
		dmap[...,ny:,:] = np.roll(imap[...,::-1], nx//2, -1)
	else:
		dmap = imap
	periodicity = np.array(dmap.shape[-2:])
	nthread = int(utils.fallback(utils.getenv("OMP_NUM_THREADS",nthread),0))
	pflat   = pixs.reshape(2,-1).T
	if epsilon is None:
		epsilon = 1e-5 if imap.dtype == np.float32 else 1e-12
	oarr    = np.zeros(imap.shape[:-2]+pflat.shape[:1], imap.dtype)
	for I in utils.nditer(imap.shape[:-2]):
		fmap = enmap.fft(dmap[I], normalize=False)
		oarr[I] = ducc0.nufft.u2nu(grid=np.asarray(fmap), coord=pflat, forward=False,
			epsilon=epsilon, nthreads=nthread, periodicity=periodicity, fft_order=True).real
		del fmap
	# Restore predims
	oarr = oarr.reshape(imap.shape[:-2]+oarr.shape[-1:])
	oarr/= dmap.npix
	return oarr

@numba.njit(nogil=True)
def rotate_pol(pmap, gamma, spin):
	if spin == 0: return
	qarr, uarr = pmap
	for i in range(qarr.shape[0]):
		for j in range(qarr.shape[1]):
			c = np.cos(spin*gamma[i,j])
			s = np.sin(spin*gamma[i,j])
			q =  qarr[i,j]*c + uarr[i,j]*s
			u = -qarr[i,j]*s + uarr[i,j]*c
			qarr[i,j] = q
			uarr[i,j] = u

def apply_modulation(map, A, T0=utils.T_cmb, freq=150e9, map_unit=1e-6, mode="T2lin",
		dipole=False, spin=[0,2]):
	if    mode in [None, "none"]: pass
	elif  mode in ["plain", "T2T"]:
		map *= A
		if dipole:
			utils.atleast_Nd(map,3)[...,0,:,:] += (A-1)*(T0/map_unit)
	elif  mode in ["T2lin", "lin2T", "lin2lin"]:
		# We're in linearized thermodynamic units. We assume that the map doesn't contain the
		# monopole, so we can treat it as a perturbation around the monopole. If the map
		# contains the monopole, then linearized units probably isn't the best choice
		for s, I in enmap.spin_pre_helper(spin, map.shape[:-2]):
			for comp in map[I]:
				if mode == "T2lin":
					_modulate_T2lin(comp, A, T0, freq, map_unit, spin=s, dipole=dipole)
				elif mode == "lin2T":
					_modulate_lin2T(comp, A, T0, freq, map_unit, spin=s, dipole=dipole)
				else:
					# This case probably isn't necessary
					_modulate_lin2T(comp, A*0+1, T0, freq, map_unit, spin=s, dipole=dipole)
					_modulate_T2lin(comp, A, T0, freq, map_unit, spin=s, dipole=dipole)
	else: raise ValueError("Urecognized modulation mode '%s'" % mode)
	return map

planck  = numba.njit(utils.planck)
dplanck = numba.njit(utils.dplanck)
iplanck = numba.njit(utils.iplanck_T)

# This is 4x as slow as the old one, but handles any value of beta, and also non-buggy :)
# This isn't the bottleneck anyway - aberration is - so I think it's worth it to just use this.
# In the future I can easily gain back af actor 4x with an implementation in C
@numba.njit
def _modulate_T2lin(map, A, T0=utils.T_cmb, freq=150e9, map_unit=1e-6, spin=0, dipole=False):
	scale= dplanck(freq, T=T0)
	off  = planck(freq,T0)/scale
	for y in range(map.shape[0]):
		for x in range(map.shape[1]):
			ival = np.float64(map[y,x])
			a    = np.float64(A[y,x])
			# This eats up at around 6 digits of precision for polarization, so
			# we need double precision
			oval = planck(freq, a*(ival*map_unit+T0))/scale
			if spin == 0 and dipole:
				oval -= off
			else:
				oval -= planck(freq, a*T0)/scale
			map[y,x] = oval/map_unit

@numba.njit
def _modulate_lin2T(map, A, T0=utils.T_cmb, freq=150e9, map_unit=1e-6, spin=0, dipole=False):
	scale= dplanck(freq, T=T0) # Jy/sr/K'
	off  = planck(freq,T0)     # Jy/sr
	for y in range(map.shape[0]):
		for x in range(map.shape[1]):
			a     = np.float64(A[y,x])
			ival  = np.float64(map[y,x]) # µK'
			# lin = P(T)/(dP/dT|T0) => T = P"(lin*(dP/dT|T0))
			oval  = ival*map_unit*scale # Jy/sr
			if spin == 0 and dipole:
				oval += off
			else:
				# Pre-emptively add a dipole that will be cancelled
				# when we apply the modulation below
				oval += planck(freq, 1/a*T0)
			# Go to full T units
			oval  = iplanck(freq, oval)
			# Apply modulation
			oval *= a
			oval -= utils.T_cmb
			map[y,x] = oval/map_unit

@numba.njit
def fast_rewind(arr, period, ref=None):
	if ref is None: ref = period/2
	ref2 = ref - period/2
	for i, a in enumerate(arr):
		# Try to make it branchless
		off     = a-ref
		arr[i] -= period*(off>=period)
		arr[i] += period*(off< 0)

# Should move this to enmap

def sky2pix(shape, wcs, pos):
	typ = wcs.wcs.ctype[0][-3:]
	if typ == "CAR" and wcs.wcs.crval[1] == 0:
		# Simple Plate Carree
		return np.array([
			(pos[0]-wcs.wcs.crval[1]*utils.degree)/(wcs.wcs.cdelt[1]*utils.degree)+(wcs.wcs.crpix[1]-1),
			(pos[1]-wcs.wcs.crval[0]*utils.degree)/(wcs.wcs.cdelt[0]*utils.degree)+(wcs.wcs.crpix[0]-1),
		])
	else:
		return enmap.sky2pix(shape, wcs, pos)



"""Module for making it a bit more convenient to work with sqlite files.
Similar to using a Connection directly, but also provides easy access
to a list of which tables and columns are available, which I always find
cumbersome to get at otherwise"""
import sqlite3, pprint, contextlib, tempfile, os

class SQL:
	def __init__(self, fname=":memory:"):
		# Is it a file name?
		if isinstance(fname, str):
			self.conn = sqlite3.connect(fname)
			self.fname= fname
			self.own  = True
		# Is it an SQL or similar?
		elif hasattr(fname, "conn"):
			self.conn = fname.conn
			self.fname= get_fname(self.conn)
			self.own  = False
		# Is it a plain sqlite3 connection?
		elif hasattr(fname, "execute"):
			self.conn  = fname
			self.fname = get_fname(self.conn)
			self.own   = False
		else: raise ValuError("SQL.__init__ needs either a file name, an SQL object or an sqlite3 connection")
	def execute(self, command, args=[]):
		return self.conn.execute(command, args)
	def executemany(self, command, args=[]):
		return self.conn.executemany(command, args)
	def derive(self, query, tname="result", aname="_src"):
		return derive(self.conn, query, tname=tname, aname=aname)
	def close(self):
		if self.own: self.conn.close()
	def backup(self, other): backup(self, other)
	def attach(self, other, name="other", mode="r"):
		"""Context manager. Temporarily attaches other to us.
		To do it permanently, just use a manual execute."""
		return attach(self, other, name=name, mode=mode)
	def tables(self): return tables(self)
	def columns(self, tname): return columns(self, tname)
	def show(self, table, limit=10): return show(self, table, limit=limit)
	def __repr__(self): return info(self, "SQL", extra=["fname='%s'" % self.fname, "own=%d" % self.own])
	def __enter__(self): return self
	def __exit__(self, *args, **kwargs):
		self.close()

def tables(conn):
	return [entry[0] for entry in conn.execute("SELECT name from sqlite_master WHERE type='table';")]
def columns(conn, tname):
	return [col[1] for col in conn.execute("PRAGMA table_info('%s');" % tname)]
def rows(conn, tname):
	return list(conn.execute("select count(*) from %s" % tname))[0][0]
def info(conn, name="Connection", extra=[]):
	tnames   = tables(conn)
	coldescs = ["%s*%d" % (str(columns(conn, tname)),rows(conn,tname)) for tname in tnames]
	tabdescs = ", ".join(["%s:%s" % (tname, coldesc) for tname, coldesc in zip(tnames, coldescs)])
	fields = ["tables=[%s]" % tabdescs]+extra
	return "%s(%s)" % (name, ", ".join(fields))
def show(conn, table, limit=10):
	query = table # either table name or a full query
	toks  = query.split()
	if len(toks) == 1:
		# Someone just gave us a table name. Rewrite to a full format
		query = "select * from " + toks[0]
	limit_included = "limit" in toks or "LIMIT" in toks
	if not limit_included:
		# Add our limit. We add one more so we can use ... to indicate
		# there's more
		query += " limit %d" % (limit+1)
	result = list(conn.execute(query))
	if limit_included: limit = len(result)
	print(format_result(result, limit=limit))

def format_result(result, limit=None):
	if limit is None: limit = len(result)
	if len(result) == 0: return "<empty>"
	trunc = len(result) > limit
	if trunc: result = result[:limit]
	# Find the max-width of each column
	widths = [0 for field in result[0]]
	for row in result:
		for fi, field in enumerate(row):
			widths[fi] = max(widths[fi], len(str(field)))
	fmt   = " ".join(["%%%ds" % w for w in widths])
	lines = [fmt % row for row in result]
	if trunc: lines.append("...")
	return "\n".join(lines)

def get_fname(conn):
	row = next(conn.execute("pragma database_list"))
	return row[2]

def backup(source, target):
	"""Simple wrapper that supports both raw connections
	and SQL objects"""
	if hasattr(source, "conn"): source = source.conn
	if hasattr(target, "conn"): target = target.conn
	source.backup(target)

@contextlib.contextmanager
def attach(conn_base, conn_other, name="other", mode="r"):
	"""Context manager to temporarily attach one database to another.
	Useful when transfering data between two databases.

	Works around sqlite's lack of support for attaching memory-only
	databases to each other by dumping other into an anonymous temporary
	file, and then restoring it afterwards if mode contains "w".
	Yuck.
	"""
	fname = get_fname(conn_other)
	if fname:
		conn_base.execute("attach database '%s' as %s" % (fname, name))
		try:
			yield
		finally:
			conn_base.execute("detach database %s" % name)
	else:
		with tempfile.NamedTemporaryFile(suffix=".sqlite") as tfile:
			with SQL(tfile.name) as tmp_db: # this ensures it's auto-closed
				if "r" in mode: backup(conn_other, tmp_db)
				conn_base.execute("attach database '%s' as %s" % (tfile.name, name))
				try:
					yield
				finally:
					conn_base.execute("detach database %s" % name)
					if "w" in mode: backup(tmp_db, conn_other)

def derive(conn, query, tname="result", out_conn=None, aname="_src"):
	"""Return a new temporary database containing the result of
	running query on the provided connection."""
	if out_conn:
		# This may seem backwards, but doing it from the old db
		# has two advantages: 1. It works if the old one is in-memory,
		# and 2. one doesn't need to modify the query
		with attach(conn, out_conn, aname, mode="rw"):
			conn.execute("create table %s.%s as %s" % (aname, tname, query))
		return out_conn
	else:
		derived = SQL()
		try:
			return derive(conn, query, tname=tname, out_conn=derived)
		except:
			# Yes, this will only happen on error. The normal close will happen
			# as part of the user using a with with our result.
			# We do it this way to support both with and manual open/close
			if derived: derived.close()
			raise

# Make sqlite.open an alias for sqlite.SQL
open = SQL

# Full list of SQLite keywords
keywords = set([
	"abort", "action", "add", "after", "all", "alter", "always", "analyze", "and", "as",
	"asc", "attach", "autoincrement", "before", "begin", "between", "by", "cascade", "case",
	"cast", "check", "collate", "column", "commit", "conflict", "constraint", "create",
	"cross", "current", "current_date", "current_time", "current_timestamp", "database",
	"default", "deferrable", "deferred", "delete", "desc", "detach", "distinct", "do",
	"drop", "each", "else", "end", "escape", "except", "exclude", "exclusive", "exists",
	"explain", "fail", "filter", "first", "following", "for", "foreign", "from", "full",
	"generated", "glob", "group", "groups", "having", "if", "ignore", "immediate", "in",
	"index", "indexed", "initially", "inner", "insert", "instead", "intersect", "into",
	"is", "isnull", "join", "key", "last", "left", "like", "limit", "match", "materialized",
	"natural", "no", "not", "nothing", "notnull", "null", "nulls", "of", "offset", "on",
	"or", "order", "others", "outer", "over", "partition", "plan", "pragma", "preceding",
	"primary", "query", "raise", "range", "recursive", "references", "regexp", "reindex",
	"release", "rename", "replace", "restrict", "returning", "right", "rollback", "row",
	"rows", "savepoint", "select", "set", "table", "temp", "temporary", "then", "ties",
	"to", "transaction", "trigger", "unbounded", "union", "unique", "update", "using",
	"vacuum", "values", "view", "virtual", "when", "where", "window", "with", "without",
])



# -*- coding: utf-8 -*-

"""Top-level package for pixell."""

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("pixell")
except PackageNotFoundError:
    __version__ = "unknown"
    
__author__ = """Simons Observatory Collaboration Analysis Library Task Force"""
__email__ = ''



# Get information about the process's memory usage. Ideally this would just
# be some simple calls to resource, but that library is almost useless. The
# python standard library is usually good, but here it's terrible.
# If we wanted to introduce an external dependency we could try psutil,
# though I would avoid it if possible.
# It looks like ctypes would work for this. It has some overhead, but this
# doesn't need to be that fast (and proc reading has overhead too)
import resource, os, ctypes

def current():  return fallback([(linux_current,  IOError), (mac_current,  AttributeError)])
def resident(): return fallback([(linux_resident, IOError), (mac_resident, AttributeError)])
def max():      return fallback([(linux_max,      IOError), (mac_max,      AttributeError)])

def fallback(things, default=lambda:0):
	for func, exceptions in things:
		try: return func()
		except exceptions: pass
	return default()

#### Linux stuff ####

def linux_current():
	with open("/proc/self/statm","r") as f:
		return int(f.readline().split()[0])*resource.getpagesize()

def linux_resident():
	with open("/proc/self/status","r") as f:
		for line in f:
			toks = line.split()
			if toks[0] == "VmRSS:":
				return int(toks[1])*1024

def linux_max():
	with open("/proc/self/status","r") as f:
		for line in f:
			toks = line.split()
			if toks[0] == "VmPeak:":
				return int(toks[1])*1024

##### MacOs stuff #####

def mac_current():  return get_mac_taskinfo().virtual_size
def mac_resident(): return get_mac_taskinfo().resident_size
def mac_max():      return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss # bytes on mac

# Taskinfo stuff for memory lookups on macs
_libc = None
def get_mac_taskinfo():
	# Cache libc so we avoid a file system search every time this is called
	global _libc
	if _libc is None: _libc = ctypes.cdll.LoadLibrary(None)
	# Fail immediately if task_info doesn't exist
	_libc.task_info
	# Define data structurs
	from ctypes import c_int, c_uint, c_ulong
	class time_value_t(ctypes.Structure):
		_fields_ = [("seconds", c_int), ("microseconds", c_int)]
	class task_basic_info(ctypes.Structure):
		_pack_ = 4
		_fields_ = [("suspend_count", c_int), ("virtual_size", c_ulong), ("resident_size", c_ulong),
				("user_time", time_value_t), ("system_time", time_value_t), ("policy", c_int)]
	count = c_uint(ctypes.sizeof(task_basic_info)//ctypes.sizeof(c_uint))
	# Define function interfaces
	task_self = _libc.mach_task_self
	task_self.restype = c_uint
	task_self.argtypes = []
	me        = task_self()
	task_info = _libc.task_info
	task_info.restype = c_int
	task_info.argtypes = [c_uint, c_uint, ctypes.POINTER(task_basic_info), ctypes.POINTER(c_uint)]
	info    = task_basic_info()
	status  = _libc.task_info(me, 5, ctypes.byref(info), ctypes.byref(count))
	return info if status == 0 else None



import numpy as np
from . import utils, _array_ops_32, _array_ops_64

# This is a reduced version of enlib.array_ops. Everything that depends
# on lapack has been removed, as has complex number support.

def get_core(dtype):
	if   dtype == np.float32:    return _array_ops_32.array_ops
	elif dtype == np.float64:    return _array_ops_64.array_ops
	raise ValueError("Unsupported data type: %s" % str(dtype))

def find_contours(imap, vals, omap=None):
	core = get_core(imap.dtype)
	if omap is None:
		omap = imap.astype(np.int32)*0
	core.find_contours(imap.T, vals, omap.T)
	return omap

def roll_rows(imap, offsets, omap=None):
	if omap is None: omap = np.zeros_like(imap)
	# Loop over outer dimensions, since fortran implementation
	# doesn't support those. This should be cheap
	for I in utils.nditer(imap.shape[:-2]):
		get_core(imap.dtype).roll_rows(imap[I].T, offsets, omap[I].T)
	return omap

def ang2rect(a):
	core = get_core(a.dtype)
	res = np.zeros([len(a),3],dtype=a.dtype)
	core.ang2rect(a.T,res.T)
	return res

def matmul(A, B, axes=[-2,-1]):
	# Massage input arrays. This should be factored out,
	# as it is common for many functions
	axes = [i if i >= 0 else A.ndim+i for i in axes]
	bax  = axes[:len(axes)-(A.ndim-B.ndim)]
	Af = utils.partial_flatten(A,axes)
	Bf = utils.partial_flatten(B,bax)
	mustadd = Bf.ndim == 2
	if mustadd: Bf = Bf[:,None,:]
	Bf = np.ascontiguousarray(Bf)
	if A.dtype != B.dtype:
		dtype = np.result_type(A.dtype,B.dtype)
		Af = Af.astype(dtype,copy=False)
		Bf = Bf.astype(dtype,copy=False)
	# Compute the shape of the output array
	Xf = np.empty((Bf.shape[0],Bf.shape[1],Af.shape[1]),dtype=Bf.dtype)
	# Actually perform the operation
	core = get_core(Bf.dtype)
	core.matmul_multi(Af.T, Bf.T, Xf.T)
	# Unwrangle
	if mustadd: Xf = Xf[:,0,:]
	X = utils.partial_expand(Xf, B.shape, bax)
	return X

def wrap_mm_m(name, vec2mat=False):
	"""Wrap a fortran subroutine which takes (n,n,m),(n,k,m) and overwrites
	its second argument to a python function where the "n" axes can be
	at arbitrary locations, specified by the axes argument, and where
	the arrays can be arbitrary-dimensional. These are all flattened
	internally. If vec2mat is specified, the second argument will have
	a dummy axis added internally if needed."""
	def f(A,B,axes=[-2,-1]):
		axes = [i if i >= 0 else A.ndim+i for i in axes]
		bax  = axes[:len(axes)-(A.ndim-B.ndim)]
		B  = B.copy()
		Af = utils.partial_flatten(A,axes)
		Bf = utils.partial_flatten(B,bax)
		mustadd = vec2mat and Bf.ndim == 2
		if mustadd: Bf = Bf[:,None,:]
		Bf = np.ascontiguousarray(Bf)
		assert A.dtype == B.dtype
		core = get_core(A.dtype)
		fun  = getattr(core, name)
		fun(Af.T, Bf.T)
		if mustadd: Bf = Bf[:,0,:]
		B[...] = utils.partial_expand(Bf, B.shape, bax)
		return B
	return f

matmul_sym   = wrap_mm_m("matmul_multi_sym", vec2mat=True)



import unittest

from pixell import enmap, wcsutils, utils
DEG = utils.degree
import numpy as np

def is_centered(pixel_index, rtol=1e-6):
    """Returns element-by-element True if pixel_index is sufficiently
    close to integer-valued."""
    frac = (np.asarray(pixel_index) + 0.5) % 1.0 - 0.5
    return np.isclose(frac, 0., rtol=rtol)

class Patch:
    ra_range = None  #! Right Ascension; left side, right side.
    dec_range = None #! Declination; bottom, top.
    @classmethod
    def centered_at(cls, ra0, dec0, width, height):
        self = cls()
        self.ra_range = np.array((ra0+width/2, ra0-width/2))
        self.dec_range = np.array((dec0-height/2, dec0+height/2))
        return self

    def pos(self):
        return np.array([[self.dec_range[0], self.ra_range[0]],
                         [self.dec_range[1], self.ra_range[1]]]) * DEG
    def extent(self, delta_ra = 0., delta_dec = 0.):
        return [self.ra_range[0]  + delta_ra /2, self.ra_range[1]  - delta_ra /2,
                self.dec_range[0] - delta_dec/2, self.dec_range[1] + delta_dec/2]
    def center(self):
        return 0.5 * np.array([(self.dec_range[0] + self.dec_range[1]),
                               (self.ra_range[0] + self.ra_range[1])])


class GeometryTests(unittest.TestCase):

    def test_reference(self):
        """Test that WCS are properly adjusted, on request, to put a reference
        pixel at integer pixel number.

        """
        DELT = 0.1
        # Note we're adding a half-pixel margin to stay away from rounding cut.
        patch = Patch.centered_at(-52., -38., 12. + DELT, 12.0 + DELT)
        # Test a few reference RAs.
        for ra1 in [0., 0.001, 90.999, 91.101, 120., 179.9, 180.0, 180.1, 270.]:
            shape, wcs = enmap.geometry(pos=patch.pos(),
                                        res=DELT*DEG,
                                        proj='cea',
                                        ref=(0, ra1*DEG))
            ref = np.array([[ra1,0]])
            ref_pix = wcs.wcs_world2pix(ref, 0)
            assert(np.all(is_centered(ref_pix)))

    def test_zenithal(self):
        DELT = 0.05
        patch0 = Patch.centered_at(308., -38., 1.+DELT, 1.+DELT)
        patch1 = Patch.centered_at(309., -39., 1.+DELT, 1.+DELT)
        ref = patch0.center()   # (dec,ra) in degrees.
        for proj in ['tan', 'zea', 'air']:
            print('Checking reference tracking of "%s"...' % proj)
            shape0, wcs0 = enmap.geometry(pos=patch0.pos(),
                                          res=DELT*utils.degree,
                                          proj=proj,
                                          ref=ref*DEG)
            shape1, wcs1 = enmap.geometry(pos=patch1.pos(),
                                          res=DELT*utils.degree,
                                          proj=proj,
                                          ref=ref*DEG)
            # Note world2pix wants [(ra,dec)], in degrees...
            pix0 = wcs0.wcs_world2pix(ref[::-1][None],0)
            pix1 = wcs1.wcs_world2pix(ref[::-1][None],0)
            print(shape0,wcs0,pix0)
            assert(np.all(is_centered(pix0)))
            assert(np.all(is_centered(pix1)))

    def test_zenithal_lonpole(self):
        """For zenithal projs, test that lonpole is set to 180, even
        if the reference point is the north pole.

        """
        DELT = 0.05
        for geometry_gen in [enmap.geometry, enmap.geometry2]:
            for ep in [DELT * 2, 0]:
                patch = Patch.centered_at(0., 90.-ep, 0., 0.)
                for proj in ['tan', 'zea', 'arc', 'sin']:
                    shape, wcs = geometry_gen(pos=patch.center() * DEG,
                                              shape=(101, 101),
                                              res=DELT*utils.degree,
                                              proj=proj)
                    dec, ra = enmap.posmap(shape, wcs)
                    # Bottom row of the map only have RA near 0 (not 180).
                    assert np.all(abs(ra[0]) < 90*DEG)

    def test_full_sky(self):
        """Test that fullsky_geometry returns sensible objects.

        Or at least the objects that we considered sensible when we
        wrote this test.

        """
        shape, w = enmap.fullsky_geometry(res=0.01*DEG, proj='car')
        ny, nx = shape
        for delta, expect_nans in [(0., False), (.001, True)]:
            for ix in [-0.5-delta,nx-0.5+delta]:
                for iy in [0-delta, ny-1+delta]:
                    c = w.wcs_pix2world([(ix,iy)], 0)
                    #print(ix,iy,c)
                    assert np.any(np.isnan(c)) == expect_nans


    def test_area(self):
        """Test that map area is computed accurately."""
        test_patches = []
        # Small CAR patch
        DELT = 0.01
        patch = Patch.centered_at(-52., -38., 12. + DELT, 12.0 + DELT)
        shape, w = enmap.geometry(pos=patch.pos(),
                                  res=DELT*DEG,
                                  proj='car',
                                  ref=(0, 0))
        exact_area = (np.dot(patch.ra_range*DEG, [-1,1]) *
                      np.dot(np.sin(patch.dec_range*DEG), [1,-1]))

        test_patches.append((shape, w, exact_area))
        # Full sky CAR patch
        shape, w = enmap.fullsky_geometry(res=0.01*DEG, proj='car')
        exact_area = 4*np.pi
        test_patches.append((shape, w, exact_area))
        # Small ZEA patch at pole
        shape, w = enmap.geometry(pos=[90*DEG,0], res=DELT*DEG, proj='zea', shape=[100,100])
        exact_area = 1*DEG**2
        test_patches.append((shape, w, exact_area))

        for shape, w, exact_area in test_patches:
            ratio = enmap.area(shape, w)/exact_area
            print(ratio)
            assert(abs(ratio-1) < 1e-6)

    def test_geom_args(self):
        """Test that the different combinations of passing and not passing arguments
        to geometry() work."""
        box = np.array([[40,100],[45,95]])*utils.degree
        res = 1*utils.degree
        def close(a,b): return np.all(np.isclose(a,b,atol=1e-6*utils.degree))
        # Plain
        shape, wcs = enmap.geometry(box, res=res, proj="car")
        assert(close(shape, [5,5]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.00000000,100.00000000]))
        shape, wcs = enmap.geometry(box, res=res, proj="cea")
        assert(close(shape, [7,5]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.17551345,100.00000000]))
        shape, wcs = enmap.geometry(box, res=res, proj="zea")
        assert(close(shape, [5,4]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.49221947, 98.81479953]))
        shape, wcs = enmap.geometry(box, res=res, proj="tan")
        assert(close(shape, [5,4]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.49336943, 98.81407198]))
        shape, wcs = enmap.geometry(box, res=res, proj="air")
        assert(close(shape, [2,10]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.42085960, 99.75799905]))
        shape, wcs = enmap.geometry(box, res=res, proj="plain")
        assert(close(shape, [5,5]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.50000000, 99.50000000]))
        # Swap box order
        shape, wcs = enmap.geometry(box[::-1], res=res, proj="car")
        assert(close(shape, [5,5]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [45.00000000, 95.00000000]))
        shape, wcs = enmap.geometry(box[::-1], res=res, proj="cea")
        assert(close(shape, [7,5]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [44.59207383, 95.00000000]))
        shape, wcs = enmap.geometry(box[::-1], res=res, proj="zea")
        assert(close(shape, [5,4]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [44.49175753, 96.09829319]))
        shape, wcs = enmap.geometry(box[::-1], res=res, proj="tan")
        assert(close(shape, [5,4]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [44.49062589, 96.09912004]))
        shape, wcs = enmap.geometry(box[::-1], res=res, proj="air")
        assert(close(shape, [2,10]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [44.88532797, 95.12307674]))
        shape, wcs = enmap.geometry(box[::-1], res=res, proj="plain")
        assert(close(shape, [5,5]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [44.50000000, 95.50000000]))
        # Specify center and shape instead of box. FIXME: These should be changed to make them right-handed.
        pos   = np.array([50,120])*utils.degree
        shape = (5,5)
        oshape, wcs = enmap.geometry(pos, res=res, shape=shape, proj="car")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [48.00000000,122.00000000]))
        oshape, wcs = enmap.geometry(pos, res=res, shape=shape, proj="cea")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [48.58804987,122.00000000]))
        oshape, wcs = enmap.geometry(pos, res=res, shape=shape, proj="zea")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [47.96024416,122.98709530]))
        oshape, wcs = enmap.geometry(pos, res=res, shape=shape, proj="tan")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [47.96213990,122.98447863]))
        oshape, wcs = enmap.geometry(pos, res=res, shape=shape, proj="plain")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [48.00000000,118.00000000]))
        # Same but with explicit 2d res
        oshape, wcs = enmap.geometry(pos, res=[res,-res], shape=shape, proj="car")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [48.00000000,122.00000000]))
        oshape, wcs = enmap.geometry(pos, res=[res,-res], shape=shape, proj="cea")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [48.58804987,122.00000000]))
        oshape, wcs = enmap.geometry(pos, res=[res,-res], shape=shape, proj="zea")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [47.96024416,122.98709530]))
        oshape, wcs = enmap.geometry(pos, res=[res,-res], shape=shape, proj="tan")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [47.96213990,122.98447863]))
        oshape, wcs = enmap.geometry(pos, res=[res,-res], shape=shape, proj="plain")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [48.00000000,122.00000000]))
        # Box and shape
        oshape, wcs = enmap.geometry(box, shape=shape, proj="car")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.00000000,100.00000000]))
        oshape, wcs = enmap.geometry(box, shape=shape, proj="cea")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.03023148,100.00000000]))
        oshape, wcs = enmap.geometry(box, shape=shape, proj="zea")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.48400209, 99.43687135]))
        oshape, wcs = enmap.geometry(box, shape=shape, proj="tan")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.48319549, 99.43761826]))
        oshape, wcs = enmap.geometry(box, shape=shape, proj="air")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.04076497,100.02606306]))
        oshape, wcs = enmap.geometry(box, shape=shape, proj="plain")
        assert(close(oshape, shape))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [40.50000000, 99.50000000]))

    def test_thumb_geom_args(self):
        """Test that the different combinations of passing and not passing arguments
        to thumbnail_geometry() work."""
        r   = 5*utils.degree
        res = 1*utils.degree
        def close(a,b): return np.all(np.isclose(a,b,atol=1e-6*utils.degree))
        shape, wcs = enmap.thumbnail_geometry(r=r, res=res, proj="car")
        assert(close(shape, [11,11]))
        #assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-5.00000000, -5.00000000]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-5.00000000,  5.00000000]))
        shape, wcs = enmap.thumbnail_geometry(r=r, res=res, proj="cea")
        assert(close(shape, [11,11]))
        #assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-5.00636804, -5.00000000  ]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-5.00636804,  5.00000000  ]))
        shape, wcs = enmap.thumbnail_geometry(r=r, res=res, proj="zea")
        assert(close(shape, [11,11]))
        #assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-4.99680325, -5.01591450]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-4.99680325,  5.01591450]))
        shape, wcs = enmap.thumbnail_geometry(r=r, res=res, proj="air")
        assert(close(shape, [11,11]))
        #assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-4.99205428, -5.01111089]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-4.99205428,  5.01111089]))
        shape, wcs = enmap.thumbnail_geometry(r=r, res=res, proj="tan")
        assert(close(shape, [11,11]))
        #assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-4.96857729, -4.98736529]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-4.96857729,  4.98736529]))
        shape, wcs = enmap.thumbnail_geometry(r=r, res=res, proj="plain")
        assert(close(shape, [11,11]))
        #assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-5.00000000, -5.00000000]))
        assert(close(enmap.pix2sky(shape, wcs, [0,0])/utils.degree, [-5.00000000, -5.00000000]))

if __name__ == '__main__':
    unittest.main()



import sys
sys.path.append('../../tests')
import test_pixell as ptests
import os
import numpy as np
from pixell import enmap

version = sys.argv[1]

obs_pos,grad,raw_pos = ptests.get_offset_result(1.,np.float64)
enmap.write_map("MM_offset_obs_pos_%s.fits" % version,obs_pos)
enmap.write_map("MM_offset_grad_%s.fits"  % version,grad)
enmap.write_map("MM_offset_raw_pos_%s.fits"  % version,raw_pos)

lensed,unlensed = ptests.get_lens_result(1.,400,np.float64)
enmap.write_map("MM_lensed_%s.fits"  % version,lensed)
enmap.write_map("MM_unlensed_%s.fits"  % version,unlensed)



"""
Core benchmark script to be evaluated with various threading
configurations. See benchmark_pixell.py for the actual benchmark
that varies the number of threads.
"""

from pixell import curvedsky, enmap, utils
import time
import numpy as np


def main():
    np.random.seed(100)
    shape, wcs = enmap.fullsky_geometry(res=12.0 * utils.arcmin)
    imap = enmap.enmap(np.random.random(shape), wcs)

    nsims = 40
    lmax = int(6000 * (2.0 / 16.0))

    t0 = time.time()
    for i in range(nsims):
        alm = curvedsky.map2alm(imap, lmax=lmax)
        curvedsky.alm2map(alm, enmap.empty(shape, wcs))
    t1 = time.time()

    total = t1 - t0
    print(f"{total:.4f} seconds.")


if __name__ == "__main__":
    main()
